import os
import urllib.request
from kdezero.utils import cache_dir


def show_progress(block_num, block_size, total_size):
    """function to display progress bar for urllib.request.urlretrieve.
    """
    bar_template = "\r[{}] {:.2f}%"

    downloaded = block_num * block_size
    p = downloaded / total_size * 100
    i = int(downloaded / total_size * 30)
    if p >= 100.0: p = 100.0
    if i >= 30: i = 30
    bar = "#" * i + "." * (30 - i)
    print(bar_template.format(bar, p), end='')


def get_file(url, file_name=None, file_dir=cache_dir):
    """Download a file from the 'url' if it is not in the cache.

    By default, the file at the 'url' is downloaded to the '~/.kdezero'.
    Args:
        url (str): URL of the file.
        file_name (str, optional):
            Name of the file. It 'None' is specified the original
            file name is used.
        file_dir (str, optional):
            Output directory.

    Returns:
        str: path to the saved file. (cache_dir / file_name)
    """
    if file_name is None:
        file_name = url[url.rfind('/') + 1:]
    file_path = os.path.join(file_dir, file_name)

    if not os.path.exists(file_dir):
        os.mkdir(file_dir)

    if os.path.exists(file_path):
        return file_path

    print("Downloading: " + file_name)
    try:
        urllib.request.urlretrieve(url, file_path, show_progress)
    except(Exception, KeyboardInterrupt) as e:
        if os.path.exists(file_path):
            os.remove(file_path)
        raise
    print(" Done")

    return file_path



import os

cache_dir = os.path.join(os.path.expanduser('~'), '.kdezero')


from kdezero.utils.calculation_graph import get_dot_graph
from kdezero.utils.calculation_graph import plot_dot_graph

from kdezero.utils.download_file import show_progress
from kdezero.utils.download_file import get_file

from kdezero.utils.numpy_utility import sum_to
from kdezero.utils.numpy_utility import reshape_sum_backward
from kdezero.utils.numpy_utility import logsumexp

from kdezero.utils.utils import get_deconv_outsize
from kdezero.utils.utils import get_conv_outsize
from kdezero.utils.utils import pair



from kdezero.functions.basic_calc_functions import add
from kdezero.functions.basic_calc_functions import sub
from kdezero.functions.basic_calc_functions import rsub
from kdezero.functions.basic_calc_functions import mul
from kdezero.functions.basic_calc_functions import div
from kdezero.functions.basic_calc_functions import rdiv
from kdezero.functions.basic_calc_functions import neg
from kdezero.functions.basic_calc_functions import pow

from kdezero.functions.mathematical_functions import sin
from kdezero.functions.mathematical_functions import cos
from kdezero.functions.mathematical_functions import tanh
from kdezero.functions.mathematical_functions import exp
from kdezero.functions.mathematical_functions import log

from kdezero.functions.ndarray_functions import reshape
from kdezero.functions.ndarray_functions import transpose
from kdezero.functions.ndarray_functions import get_item
from kdezero.functions.ndarray_functions import sum
from kdezero.functions.ndarray_functions import sum_to
from kdezero.functions.ndarray_functions import broadcast_to
from kdezero.functions.ndarray_functions import matmul

from kdezero.functions.activation_functions import sigmoid
from kdezero.functions.activation_functions import relu
from kdezero.functions.activation_functions import softmax

from kdezero.functions.loss_functions import mean_squared_error
from kdezero.functions.loss_functions import softmax_cross_entropy

from kdezero.functions.utils import linear
from kdezero.functions.utils import accuracy
from kdezero.functions.utils import dropout

from kdezero.functions.conv_functions import conv2d
from kdezero.functions.conv_functions import deconv2d
from kdezero.functions.conv_functions import im2col
from kdezero.functions.conv_functions import col2im
from kdezero.functions.conv_functions import pooling



import os
import subprocess
from kdezero.utils import cache_dir


def _dot_var(v, verbose=False):
    dot_var = '{} [label="{}", color=orange, style=filled]\n'

    name = '' if v.name is None else v.name
    if verbose and v.data is not None:
        if v.name is not None:
            name += ': '
        name += str(v.shape) + ' ' + str(v.dtype)
    return dot_var.format(id(v), name)


def _dot_func(f):
    dot_func = '{} [label="{}", color=lightblue, style=filled, shape=box]\n'
    txt = dot_func.format(id(f), f.__class__.__name__)

    dot_edge = '{} -> {}\n'
    for x in f.inputs:
        txt += dot_edge.format(id(x), id(f))
    for y in f.outputs:
        txt += dot_edge.format(id(f), id(y()))
    return txt


def get_dot_graph(output, verbose=True):
    """From the variables, create text that describes the structure of the model,
    written in graphviz-compatible dot language.

    Args:
        output (kdezero.Variable):
            Variables that retroactively create graphs with backpropagation.
        verbose (bool, optional):
            Whether to display in more detail. (default: True)

    Returns:
        string: dot text
    """
    txt = ''
    funcs = []
    seen_set = set()

    def add_func(f):
        if f not in seen_set:
            funcs.append(f)
            seen_set.add(f)

    add_func(output.creator)
    txt += _dot_var(output, verbose)

    while funcs:
        func = funcs.pop()
        txt += _dot_func(func)
        for x in func.inputs:
            txt += _dot_var(x, verbose)

            if x.creator is not None:
                add_func(x.creator)
    return 'digraph g {\n' + txt + '}'


def plot_dot_graph(output, verbose=True, to_file='graph.png'):
    """From the variables,
    use graphviz to create an image that represents the composition of the model.

    Args:
        output (kdezero.Variable):
            Variables that retroactively create graphs with backpropagation.
        verbose (bool, optional):
            Whether to display in more detail. (default: True)
        to_file (string, optional):
            File name to output. (default: 'graph.png')

    Note:
        Display the image if it can be displayed using a notebook etc.
    """
    dot_graph = get_dot_graph(output, verbose)

    if not os.path.exists(cache_dir):
        os.mkdir(cache_dir)
    graph_path = os.path.join(cache_dir, 'tmp_graph.dot')

    with open(graph_path, 'w') as f:
        f.write(dot_graph)

    extension = os.path.splitext(to_file)[1][1:]
    cmd = 'dot {} -T {} -o {}'.format(graph_path, extension, to_file)
    subprocess.run(cmd, shell=True)

    try:
        from IPython import display
        return display.Image(filename=to_file)
    except:
        pass



import numpy as np
from kdezero import cuda


def sum_to(x, shape):
    """Sum elements along axes to output an array of a given shape.
    Args:
        x (ndarray): Input array.
        shape:
    Returns:
        ndarray: Output array of the shape.
    """
    ndim = len(shape)
    lead = x.ndim - ndim
    lead_axis = tuple(range(lead))

    axis = tuple([i + lead for i, sx in enumerate(shape) if sx == 1])
    y = x.sum(lead_axis + axis, keepdims = True)
    if lead > 0:
        y = y.squeeze(lead_axis)
    return y


def reshape_sum_backward(gy, x_shape, axis, keepdims):
    """Reshape gradient appropriately for kdezero.functions.sum's backward.

    Args:
        gy (kdezero.Variable):
            Gradient variable from the output by backprop.
        x_shape (tuple):
            Shape used at sum function's forward.
        axis (None or int or tuple of ints):
            Axis used at sum function's forward.
        keepdims (bool):
            Keepdims used at sum function's forward.
    Returns:
        kdezero.Variable: Gradient variable which is reshaped appropriately
    """
    ndim = len(x_shape)
    tupled_axis = axis
    if axis is None:
        tupled_axis = None
    elif not isinstance(axis, tuple):
        tupled_axis = (axis,)
    
    if not (ndim == 0 or tupled_axis is None or keepdims):
        actual_axis = [a if a >= 0 else a + ndim for a in tupled_axis]
        shape = list(gy.shape)
        for a in sorted(actual_axis):
            shape.insert(a, 1)
    else:
        shape = gy.shape

    gy = gy.reshape(shape)
    return gy


def logsumexp(x, axis=1):
    """After exp, sum elements along axes, and apply log.

    Args:
        x (ndarray):
        axis (int, optional): (default: 1)

    Returns:
        ndarray:
    """
    xp = cuda.get_array_module(x)
    m = x.max(axis=axis, keepdims=True)
    y = x - m
    xp.exp(y, out=y)
    s = y.sum(axis=axis, keepdims=True)
    xp.log(s, out=s)
    m += s
    return m



import numpy as np
from kdezero.datasets import Dataset


def get_spiral(train=True):
    seed = 1984 if train else 2020
    np.random.seed(seed=seed)

    num_data, num_class, input_dim = 100, 3, 2
    data_size = num_class * num_data
    x = np.zeros((data_size, input_dim), dtype=np.float32)
    t = np.zeros(data_size, dtype=np.int)

    for j in range(num_class):
        for i in range(num_data):
            rate = i / num_data
            radius = 1.0 * rate
            theta = j * 4.0 + 4.0 * rate + np.random.randn() * 0.2
            ix = num_data * j + i
            x[ix] = np.array([radius * np.sin(theta),
                              radius * np.cos(theta)]).flatten()
            t[ix] = j
    indices = np.random.permutation(num_data * num_class)
    x = x[indices]
    t = t[indices]
    return x, t


class Spiral(Dataset):
    """Spiral data for classification problems.
    The number of data is 100.
    The number of class is 3.
    And the number of dimensions is 2.

    Attribute:
        Attribute:
        data (ndarray): shape is (300, 2)
        label (ndarray): shape is (300,)
        train (bool):
            Flag for learning.
            The internal random number is changed depending on whether train is True or not.
    """
    def prepare(self):
        self.data, self.label = get_spiral(self.train)



import kdezero
from kdezero import Function
from kdezero import as_array


class Add(Function):
    def forward(self, x0, x1):
        self.x0_shape, self.x1_shape = x0.shape, x1.shape
        y = x0 + x1
        return y

    def backward(self, gy):
        gx0, gx1 = gy, gy
        if self.x0_shape != self.x1_shape:
            gx0 = kdezero.functions.sum_to(gx0, self.x0_shape)
            gx1 = kdezero.functions.sum_to(gx1, self.x1_shape)
        return gx0, gx1


def add(x0, x1):
    x1 = as_array(x1, kdezero.cuda.get_array_module(x0.data))
    return Add()(x0, x1)


class Mul(Function):
    def forward(self, x0, x1):
        return x0 * x1

    def backward(self, gy):
        x0, x1 = self.inputs
        gx0 = gy * x1
        gx1 = gy * x0
        if x0.shape != x1.shape:
            gx0 = kdezero.functions.sum_to(gx0, x0.shape)
            gx1 = kdezero.functions.sum_to(gx1, x1.shape)
        return gx0, gx1


def mul(x0, x1):
    x1 = as_array(x1, kdezero.cuda.get_array_module(x0.data))
    return Mul()(x0, x1)


class Neg(Function):
    def forward(self, x):
        return -x

    def backward(self, gy):
        return -gy


def neg(x):
    return Neg()(x)


class Sub(Function):
    def forward(self, x0, x1):
        self.x0_shape, self.x1_shape = x0.shape, x1.shape
        return x0 - x1

    def backward(self, gy):
        gx0 = gy
        gx1 = -gy
        if self.x0_shape != self.x1_shape:
            gx0 = kdezero.functions.sum_to(gx0, self.x0_shape)
            gx1 = kdezero.functions.sum_to(gx1, self.x1_shape)
        return gx0, gx1


def sub(x0, x1):
    x1 = as_array(x1, kdezero.cuda.get_array_module(x0.data))
    return Sub()(x0, x1)


def rsub(x0, x1):
    x1 = as_array(x1, kdezero.cuda.get_array_module(x0.data))
    return Sub()(x1, x0)


class Div(Function):
    def forward(self, x0, x1):
        return x0 / x1

    def backward(self, gy):
        x0, x1 = self.inputs
        gx0 = gy / x1
        gx1 = gy * (-x0 / x1 ** 2)
        if x0.shape != x1.shape:
            gx0 = kdezero.functions.sum_to(gx0, x0.shape)
            gx1 = kdezero.functions.sum_to(gx1, x1.shape)
        return gx0, gx1


def div(x0, x1):
    x1 = as_array(x1, kdezero.cuda.get_array_module(x0.data))
    return Div()(x0, x1)


def rdiv(x0, x1):
    x1 = as_array(x1, kdezero.cuda.get_array_module(x0.data))
    return Div()(x1, x0)


class Pow(Function):
    def __init__(self, c):
        self.c = c

    def forward(self, x):
        return x ** self.c

    def backward(self, gy):
        x = self.inputs[0]
        c = self.c
        gx = c * x ** (c - 1) * gy
        return gx


def pow(x, c):
    return Pow(c)(x)



import numpy as np
from kdezero.datasets import Dataset


class SinCurve(Dataset):
    """The data is a sine curve, and the label advances one index by one.
    The number of data is 1000.

    Attribute:
        Attribute:
        data (ndarray):
        label (ndarray):
        train (bool):
            Flag for learning.
            If train is True, data is sin and noise.
            If not, data is cos.
    Examples:
        >>> print(dataset.data[:5])
            [[-0.04955855]
            [ 0.03048039]
            [-0.01378722]
            [-0.02327317]
            [ 0.04658464]]

        >>> print(dataset.label[:5])
            [[ 0.03048039]
            [-0.01378722]
            [-0.02327317]
            [ 0.04658464]
            [ 0.02806842]]
    """
    def prepare(self):
        num_data = 1000
        dtype = np.float64

        x = np.linspace(0, 2 * np.pi, num_data)
        noise_range = (-0.05, 0.05)
        noise = np.random.uniform(noise_range[0], noise_range[1], size=x.shape)
        if self.train:
            y = np.sin(x) + noise
        else:
            y = np.cos(x)
        y = y.astype(dtype)
        self.data = y[:-1][:, np.newaxis]
        self.label = y[1:][:, np.newaxis]



import kdezero
from kdezero import as_variable, as_array, Variable, Function
from kdezero import cuda
from kdezero.functions.ndarray_functions import matmul, sum_to


class Linear(Function):
    def forward(self, x, W, b):
        y = x.dot(W)
        if b is not None:
            y += b
        return y

    def backward(self, gy):
        x, W, b = self.inputs
        gb = None if b.data is None else sum_to(gy, b.shape)
        gx = matmul(gy, W.T)
        gW = matmul(x.T, gy)
        return gx, gW, gb


def linear(x, W, b=None):
    return Linear()(x, W, b)


def accuracy(y, t):
    y, t = as_variable(y), as_variable(t)

    pred = y.data.argmax(axis=1).reshape(t.shape)
    result = (pred == t.data)
    acc = result.mean()
    return Variable(as_array(acc))


def dropout(x, dropout_ratio=0.5):
    x = as_variable(x)

    if kdezero.Config.train:
        xp = cuda.get_array_module(x)
        mask = xp.random.rand(*x.shape) > dropout_ratio
        scale = xp.array(1.0 - dropout_ratio).astype(x.dtype)
        y = x * mask / scale
        return y
    else:
        return x



import numpy as np
from kdezero import Function
from kdezero import utils, cuda
from kdezero.functions.ndarray_functions import broadcast_to
from kdezero.functions.activation_functions import softmax


class MeanSquaredError(Function):
    def forward(self, x0, x1):
        diff = x0 - x1
        y = (diff ** 2).sum() / len(diff)
        return y

    def backward(self, gy):
        x0, x1 = self.inputs
        diff = x0 - x1
        gy = broadcast_to(gy, diff.shape)
        gx0 = gy * diff * (2. / len(diff))
        gx1 = -gx0
        return gx0, gx1


def mean_squared_error(x0, x1):
    return MeanSquaredError()(x0, x1)


class SoftmaxCrossEntropy(Function):
    def forward(self, x, t):
        N = x.shape[0]
        log_z = utils.logsumexp(x, axis=1)
        log_p = x - log_z
        log_p = log_p[np.arange(N), t.ravel()]
        y = -log_p.sum() / np.float32(N)
        return y

    def backward(self, gy):
        x, t = self.inputs
        N, CLS_NUM = x.shape

        gy *= 1/N
        y = softmax(x)
        xp = cuda.get_array_module(t.data)
        t_onehot = xp.eye(CLS_NUM, dtype=t.dtype)[t.data]
        y = (y - t_onehot) * gy
        return y


def softmax_cross_entropy(x, t):
    return SoftmaxCrossEntropy()(x, t)



from kdezero import cuda, Function


class Sin(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = xp.sin(x)
        return y

    def backward(self, gy):
        x, = self.inputs
        gx = gy * cos(x)
        return gx


def sin(x):
    return Sin()(x)


class Cos(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = xp.cos(x)
        return y

    def backward(self, gy):
        x, = self.inputs
        gx = gy * -sin(x)
        return gx


def cos(x):
    return Cos()(x)


class Tanh(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = xp.tanh(x)
        return y

    def backward(self, gy):
        y = self.outputs[0]()
        gx = gy * (1 - y * y)
        return gx


def tanh(x):
    return Tanh()(x)


class Exp(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = xp.exp(x)
        return y

    def backward(self, gy):
        y = self.outputs[0]()
        gx = gy * y
        return gx


def exp(x):
    return Exp()(x)


class Log(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = xp.log(x)
        return y

    def backward(self, gy):
        x, = self.inputs
        gx = gy / x
        return gx


def log(x):
    return Log()(x)



import numpy as np
import kdezero.layers as L
import kdezero.functions as F
from kdezero import utils


class Model(L.Layer):
    def plot(self, *inputs, to_file='model.png'):
        y = self.forward(*inputs)
        return utils.plot_dot_graph(y, verbose=True, to_file=to_file)


class MLP(Model):
    def __init__(self, fc_output_sizes, activation=F.sigmoid):
        super().__init__()
        self.activation = activation
        self.layers = []

        for i, out_size in enumerate(fc_output_sizes):
            layer = L.Linear(out_size)
            setattr(self, 'l' + str(i), layer)
            self.layers.append(layer)

    def forward(self, x):
        for l in self.layers[:-1]:
            x = self.activation(l(x))
        return self.layers[-1](x)


class VGG16(Model):
    WEIGHTS_PATH = 'https://github.com/koki0702/dezero-models/releases/download/v0.1/vgg16.npz'

    def __init__(self, pretraind=False):
        super().__init__()
        self.conv1_1 = L.Conv2d(64, kernel_size=3, stride=1, pad=1)
        self.conv1_2 = L.Conv2d(64, kernel_size=3, stride=1, pad=1)
        self.conv2_1 = L.Conv2d(128, kernel_size=3, stride=1, pad=1)
        self.conv2_2 = L.Conv2d(128, kernel_size=3, stride=1, pad=1)
        self.conv3_1 = L.Conv2d(256, kernel_size=3, stride=1, pad=1)
        self.conv3_2 = L.Conv2d(256, kernel_size=3, stride=1, pad=1)
        self.conv3_3 = L.Conv2d(256, kernel_size=3, stride=1, pad=1)
        self.conv4_1 = L.Conv2d(512, kernel_size=3, stride=1, pad=1)
        self.conv4_2 = L.Conv2d(512, kernel_size=3, stride=1, pad=1)
        self.conv4_3 = L.Conv2d(512, kernel_size=3, stride=1, pad=1)
        self.conv5_1 = L.Conv2d(512, kernel_size=3, stride=1, pad=1)
        self.conv5_2 = L.Conv2d(512, kernel_size=3, stride=1, pad=1)
        self.conv5_3 = L.Conv2d(512, kernel_size=3, stride=1, pad=1)
        self.fc6 = L.Linear(4069)
        self.fc7 = L.Linear(4069)
        self.fc8 = L.Linear(1000)

        if pretraind:
            weights_path = utils.get_file(VGG16.WEIGHTS_PATH)
            self.load_weights(weights_path)

    def forward(self, x):
        x = F.relu(self.conv1_1(x))
        x = F.relu(self.conv1_2(x))
        x = F.pooling(x, 2, 2)
        x = F.relu(self.conv2_1(x))
        x = F.relu(self.conv2_2(x))
        x = F.pooling(x, 2, 2)
        x = F.relu(self.conv3_1(x))
        x = F.relu(self.conv3_2(x))
        x = F.relu(self.conv3_3(x))
        x = F.pooling(x, 2, 2)
        x = F.relu(self.conv4_1(x))
        x = F.relu(self.conv4_2(x))
        x = F.relu(self.conv4_3(x))
        x = F.pooling(x, 2, 2)
        x = F.relu(self.conv5_1(x))
        x = F.relu(self.conv5_2(x))
        x = F.relu(self.conv5_3(x))
        x = F.pooling(x, 2, 2)
        x = F.reshape(x, (x.shape[0], -1))
        x = F.dropout(F.relu(self.fc6(x)))
        x = F.dropout(F.relu(self.fc7(x)))
        x = self.fc8(x)
        return x

    @staticmethod
    def preprocess(image, size=(224, 224), dtype=np.float32):
        image = image.convert('RGB')
        if size:
            image = image.resize(size)
        image = np.asarray(image, dtype=dtype)
        image = image[:, :, ::-1]
        image -= np.asarray([103.939, 116.779, 123.68], dtype=dtype)
        image = image.transpose((2, 0, 1))
        return image



import numpy as np
import gzip
import matplotlib.pyplot as plt
from kdezero.datasets import Dataset
from kdezero.transforms import Compose, Flatten, ToFloat, Normalize
from kdezero.utils import get_file


class MNIST(Dataset):
    """MNISET dataset
    A class that creates a MNIST dataset by downloading data
    from the URL written in the prepare function.

    Data is stored in '.kdezero' in your home directory. (~/.kdezero)

    Attribute:
        data (ndarray):
            Shape is (n, 1, 28, 28) (n, channel, height, width).
            If train == True, n == 60000. If not, n == 10000
        label (ndarray): shape is (n) and element is int.
        train (bool): Flag for learning.
        transform (Callable or None, optional):
            Object to convert data.
            By default, flatten and to float and normalize (mean is 0, std is 255)
        target_transform (Callable, optional):
            Object to convert label. (default: None)
    """
    def __init__(self,
                 train=True,
                 transform=Compose([Flatten(), ToFloat(), Normalize(0., 255.)]),
                 target_transform=None):
        super().__init__(train, transform, target_transform)

    def prepare(self):
        """Download the data and prepare the MNIST dataset.
        """
        url = 'http://yann.lecun.com/exdb/mnist/'
        train_files = {'target': 'train-images-idx3-ubyte.gz',
                       'label': 'train-labels-idx1-ubyte.gz'}
        test_files = {'target': 't10k-images-idx3-ubyte.gz',
                      'label': 't10k-labels-idx1-ubyte.gz'}
        files = train_files if self.train else test_files
        data_path = get_file(url + files['target'])
        label_path = get_file(url + files['label'])

        self.data = self._load_data(data_path)
        self.label = self._load_label(label_path)

    def _load_label(self, filepath):
        with gzip.open(filepath, 'rb') as f:
            labels = np.frombuffer(f.read(), np.uint8, offset=8)
        return labels

    def _load_data(self, filepath):
        with gzip.open(filepath, 'rb') as f:
            data = np.frombuffer(f.read(), np.uint8, offset=16)
        data = data.reshape(-1, 1, 28, 28)
        return data

    def show(self, row=10, col=10):
        """Display data randomly

        Specify the width and height, then display the data randomly for that number.

        Args:
            row (int): Number of images side by side
            col (int): Number of images arranged vertically

        Note:
            The image shall be flatten processed.
        """
        H, W = 28, 28
        img = np.zeros((H * row, W * col))
        for r in range(row):
            for c in range(col):
                img[r * H: (r + 1) * H, c * W: (c + 1) * W] = self.data[
                    np.random.randint(0, len(self.data) - 1)
                ].reshape(H, W)
        plt.imshow(img, cmap='gray', interpolation='nearest')
        plt.axis('off')
        plt.show()

    def one_show(self, index):
        """Display image by specifying index

        Args:
            index (int): Index of data.
        """
        H, W = 28, 28
        img = self.data[index].reshape(H, W)
        plt.imshow(img, cmap='gray')
        plt.axis('off')
        plt.show()

    @staticmethod
    def labels():
        return {0: '0', 1: '1', 2: '2', 3: '3', 4: '4', 5: '5', 6: '6', 7: '7', 8: '8', 9: '9'}



import numpy as np


class Dataset:
    """Class that stores data.
    This class can also preprocess the data.

    Attribute:
        train (bool, optional): Flag for learning. (default: True)
        transform (Callable or None, optional):
            Object to convert data. (default: None)
        target_transform (Callable, optional):
            Object to convert label. (default: None)
        data (ndarray or list):
            Set data in the inherited class.
        label (ndarray or list or None):
            Set label in the inherited class.
    """
    def __init__(self, train=True, transform=None, target_transform=None):
        self.train = train
        self.transform = transform
        self.target_transform = target_transform
        if self.transform is None:
            self.transform = lambda x: x
        if self.target_transform is None:
            self.target_transform = lambda x: x

        self.data = None
        self.label = None
        self.prepare()

    def __getitem__(self, index):
        """returns data and label

        Returns:
            tuple of data and label.
            If there is no label, tuple of data and None.
        """
        assert np.isscalar(index)
        if self.label is None:
            return self.transform(self.data[index]), None
        else:
            return self.transform(self.data[index]), \
                   self.target_transform(self.label[index])

    def __len__(self):
        return len(self.data)

    def prepare(self):
        pass



import numpy as np

gpu_enable = True
try:
    import cupy as cp
    cupy = cp
except ImportError:
    gpu_enable = False
from kdezero import Variable


def get_array_module(x):
    """Returns the array module for 'x'.

    Args:
        x (kdezero.Variable or numpy.ndarray or cupy.ndarray):
            Values to determine whether NumPy or CuPy should be used.

    Returns:
        module: 'cupy' or 'numpy' is returned based on the argument.
    """
    if isinstance(x, Variable):
        x = x.data

    if not gpu_enable:
        return np
    xp = cp.get_array_module(x)
    return xp


def as_numpy(x):
    """Convert to numpy.ndarray.

    Args:
        x (numpy.ndarray or cupy.ndarray):
            Arbitrary object that can be converted to numpy.ndarray.

    Returns:
        numpy.ndarray: Converted array.
    """
    if isinstance(x, Variable):
        x = x.data

    if np.isscalar(x):
        return np.array(x)
    elif isinstance(x, np.ndarray):
        return x
    return cp.asnumpy(x)


def as_cupy(x):
    """Convert to cupy.ndarray.

    Args:
        x (numpy.ndarray or cupy.ndarray):
            Arbitrary object that can be converted to cupy.ndarray.

    Returns:
        cupy.ndarray: Converted array.
    """
    if isinstance(x, Variable):
        x = x.data

    if not gpu_enable:
        raise Exception('Cupy cannot be loaded. Install Cupy!')
    return cp.asarray(x)



from kdezero import Function, as_variable
from kdezero import cuda, utils


class Reshape(Function):
    def __init__(self, shape):
        self.shape = shape

    def forward(self, x):
        self.x_shape = x.shape
        y = x.reshape(self.shape)
        return y

    def backward(self, gy):
        return reshape(gy, self.x_shape)


def reshape(x, shape):
    if x.shape == shape:
        return as_variable(x)
    return Reshape(shape)(x)


class Transpose(Function):
    def __init__(self, axes=None):
        self.axes = axes

    def forward(self, x):
        y = x.transpose(self.axes)
        return y

    def backward(self, gy):
        if self.axes is None:
            return transpose(gy)

        axes_len = len(self.axes)
        inv_axes = tuple(np.argsort([ax % axes_len for ax in self.axes]))
        return transpose(gy, inv_axes)


def transpose(x, axes=None):
    return Transpose(axes)(x)


class GetItem(Function):
    def __init__(self, slices):
        self.slices = slices

    def forward(self, x):
        y = x[self.slices]
        return y

    def backward(self, gy):
        x, = self.inputs
        f = GetItemGrad(self.slices, x.shape)
        return f(gy)


class GetItemGrad(Function):
    def __init__(self, slices, in_shape):
        self.slices = slices
        self.in_shape = in_shape

    def forward(self, gy):
        xp = cuda.get_array_module(gy)
        gx = xp.zeros(self.in_shape, dtype=gy.dtype)

        if xp is np:
            np.add.at(gx, self.slices, gy)
        else:
            xp.scatter_add(gx, self.slices, gy)
        return gx

    def backward(self, ggx):
        return get_item(ggx, self.slices)


def get_item(x, slices):
    return GetItem(slices)(x)


class Sum(Function):
    def __init__(self, axis, keepdims):
        self.axis = axis
        self.keepdims = keepdims

    def forward(self, x):
        self.x_shape = x.shape
        y = x.sum(axis=self.axis, keepdims=self.keepdims)
        return y

    def backward(self, gy):
        gy = utils.reshape_sum_backward(gy, self.x_shape, self.axis, self.keepdims)
        gx = broadcast_to(gy, self.x_shape)
        return gx


def sum(x, axis=None, keepdims=False):
    return Sum(axis, keepdims)(x)


class SumTo(Function):
    def __init__(self, shape):
        self.shape = shape

    def forward(self, x):
        self.x_shape = x.shape
        y = utils.sum_to(x, self.shape)
        return y

    def backward(self, gy):
        gx = broadcast_to(gy, self.x_shape)
        return gx


def sum_to(x, shape):
    if x.shape == shape:
        return as_variable(x)
    return SumTo(shape)(x)


class BroadcastTo(Function):
    def __init__(self, shape):
        self.shape = shape

    def forward(self, x):
        self.x_shape = x.shape
        xp = cuda.get_array_module(x)
        y = xp.broadcast_to(x, self.shape)
        return y

    def backward(self, gy):
        gx = sum_to(gy, self.x_shape)
        return gx


def broadcast_to(x, shape):
    if x.shape == shape:
        return as_variable(x)
    return BroadcastTo(shape)(x)


class Matmul(Function):
    def forward(self, x, W):
        y = x.dot(W)
        return y

    def backward(self, gy):
        x, W = self.inputs
        gx = matmul(gy, W.T)
        gW = matmul(x.T, gy)
        return gx, gW


def matmul(x, W):
    return Matmul()(x, W)



import numpy as np
import weakref
import contextlib
import kdezero

try:
    import cupy
    array_types = (np.ndarray, cupy.ndarray)
except ImportError:
    array_types = (np.ndarray)


class Config:
    """kdezero config class

    Attributes:
        enable_backprop (bool):
            Decide if you want back propagation (default: True)
        train (bool):
            Flag of whether you are training (default: True)
    """
    enable_backprop = True
    train = True


@contextlib.contextmanager
def using_config(name, value):
    """Change Config

    Can be used with 'with' syntax to temporarily change Config

    Args:
        name (string): Attribute of Config
        value: Config value to change temporarily

    """
    old_value = getattr(Config, name)
    setattr(Config, name, value)
    try:
        yield
    finally:
        setattr(Config, name, old_value)


def no_grad():
    """Turn off back propagation

    Can be used with 'with' syntax to temporarily turn off back propagation

    """
    return using_config('enable_backprop', False)


def test_mode():
    """Switch to test mode

    Can be used with 'with' syntax to temporarily
    switch from training mode to test mode

    """
    return using_config('train', False)


class Variable:
    """Variable of kdezero

    A kdezero variable with numpy format data and back propagation value

    Attributes:
        data (ndarray or None):
            value of variable
        name (string or None):
            name of variable (default: None)
        grad (kdezero.Variable or None):
            Derivative value obtained by back propagation
        creator (Function or None):
            In the case of a variable created by a function, that function is stored
        generation (int):
            Represent the generation of functions and generated variables

    Args:
        data (ndarray or None):
            value of variable
        name (string or None, optional):
            name of variable (default: None)

    Raises:
        TypeError:
            Exception occurs when the data is not the corresponding ndarray class.
    """
    __array_priority__ = 200

    def __init__(self, data, name=None):
        if data is not None:
            if not isinstance(data, array_types):
                raise TypeError('{} is not supported'.format(type(data)))

        self.data = data
        self.name = name
        self.grad = None
        self.creator = None
        self.generation = 0

    @property
    def shape(self):
        """Shape of the data as ndarray
        """
        return self.data.shape

    @property
    def ndim(self):
        """Dimension of the data as ndarray
        """
        return self.data.ndim

    @property
    def size(self):
        """Size of the data as ndarray
        """
        return self.data.size

    @property
    def dtype(self):
        """Dtype of the data as ndarray
        """
        return self.data.dtype

    def __len__(self):
        """Length of the data as ndarray
        """
        return len(self.data)

    def __repr__(self):
        if self.data is None:
            return 'variable(None)'
        p = str(self.data).replace('\n', '\n' + ' ' * 9)
        return 'variable(' + p + ')'

    def set_creator(self, func):
        """Set the function that generates this variable and set the generation

        Args:
            func (function): Function to generate this variable
        """
        self.creator = func
        self.generation = func.generation + 1

    def unchain(self):
        """Blocking backpropagation

        Set creator to None to prevent backpropagation
        """
        self.creator = None

    def cleargrad(self):
        """Reset the differential value
        """
        self.grad = None

    def backward(self, retain_grad=False, create_graph=False):
        """backpropagation

        Backpropagate retroactively to the functions
        and variables used to generate this variable.

        Args:
            retain_grad (bool):
                If it is Flase, delete all,
                leaving only the derivative value at the end. (default: False)
            create_graph (bool):
                Set to True if you want to calculate the derivative value
                during backpropagation. (default: False)
        """
        if self.grad is None:
            xp = kdezero.cuda.get_array_module(self.data)
            self.grad = Variable(xp.ones_like(self.data))

        funcs = []
        seen_set = set()

        def add_func(f):
            if f not in seen_set:
                funcs.append(f)
                seen_set.add(f)
                funcs.sort(key=lambda x: x.generation)

        add_func(self.creator)

        while funcs:
            f = funcs.pop()
            gys = [output().grad for output in f.outputs]

            with using_config('enable_backprop', create_graph):
                gxs = f.backward(*gys)
                if not isinstance(gxs, tuple):
                    gxs = (gxs,)

                for x, gx in zip(f.inputs, gxs):
                    if x.grad is None:
                        x.grad = gx
                    else:
                        x.grad = x.grad + gx
                    if x.creator is not None:
                        add_func(x.creator)

            if not retain_grad:
                for y in f.outputs:
                    y().grad = None

    def unchain_backward(self):
        """Go back and prevent backpropagation

        Go back and set creator to None to prevent backpropagation

        """
        if self.creator is not None:
            funcs = [self.creator]
            while funcs:
                f = funcs.pop()
                for x in f.inputs:
                    if x.creator is not None:
                        funcs.append(x.creator)
                        x.unchain()

    def reshape(self, *shape):
        """Reshape the data (supports back propagation)

        Args:
            shape (tuple or list of int, n ints):
                Shape after conversion

        Returns:
            kdezero.Variable: Shape changed variable
        """
        if len(shape) == 1 and isinstance(shape[0], (tuple, list)):
            shape = shape[0]
        return kdezero.functions.reshape(self, shape)

    def transpose(self, *axes):
        """Perform transpose (supports back propagation)

        Args:
            axes (tuple or list of int, None or n ints):
                Specify in the same way as transpose of ndarray.
                If None is specified or not specified, perform normal transpose.

        Returns:
            kdezero.Variable: The calculation result.
        """
        if len(axes) == 0:
            axes = None
        elif len(axes) == 1:
            if isinstance(axes[0], (tuple, list)) or axes[0] is None:
                axes = axes[0]
        return kdezero.functions.transpose(self, axes)

    @property
    def T(self):
        """normal transpose.
        """
        return kdezero.functions.transpose(self)

    def sum(self, axis=None, keepdims=False):
        """Caluculate the sum of the elements of the data (supports back propagation)

        Args:
            axis (None or int or tuple of ints, optional):
                Specify like ndarray.
            keepdims (bool, optional):
                If True, the sum is calculated while preserving the dimensions.

        Returns:
            kdezero.Variable: The calculation result.
        """
        return kdezero.functions.sum(self, axis, keepdims)

    def to_cpu(self):
        """Convert data for cpu
        """
        if self.data is not None:
            self.data = kdezero.cuda.as_numpy(self.data)

    def to_gpu(self):
        """Convert data for gpu
        """
        if self.data is not None:
            self.data = kdezero.cuda.as_cupy(self.data)


def setup_variable():
    Variable.__add__ = kdezero.functions.add
    Variable.__radd__ = kdezero.functions.add
    Variable.__mul__ = kdezero.functions.mul
    Variable.__rmul__ = kdezero.functions.mul
    Variable.__neg__ = kdezero.functions.neg
    Variable.__sub__ = kdezero.functions.sub
    Variable.__rsub__ = kdezero.functions.rsub
    Variable.__truediv__ = kdezero.functions.div
    Variable.__rtruediv__ = kdezero.functions.rdiv
    Variable.__pow__ = kdezero.functions.pow
    Variable.__getitem__ = kdezero.functions.get_item

    Variable.matmul = kdezero.functions.matmul
    Variable.dot = kdezero.functions.matmul


class Parameter(Variable):
    """Parameters calculated by machine learning
    """
    pass


class Function:
    """Function of kdezero

    Backpropagating function that can be used like numpy.

    Attribute:
        generation (int):
            Represent the generation of functions and generated variables
        inputs (list of kdezero.Variable):
            List of variables used to enter the function.
        outputs (list of kdezero.Variable):
            List of function return values.
    """
    def __call__(self, *inputs):
        """Calculation and creation of calculation graph.

        If Config.enable_backprop Flase,
        the calculation graph will not be created.

        Returns:
            kdezero.Variable or list of kdezero.Variable:
                Returns the return value of the function.
                If there are two or more return values, return a list.

        """
        inputs = [as_variable(x) for x in inputs]
        xs = [x.data for x in inputs]
        ys = self.forward(*xs)
        if not isinstance(ys, tuple):
            ys = (ys,)
        outputs = [Variable(as_array(y)) for y in ys]

        if Config.enable_backprop:
            self.generation = max([x.generation for x in inputs])
            for output in outputs:
                output.set_creator(self)
            self.inputs = inputs
            self.outputs = [weakref.ref(output) for output in outputs]

        return outputs if len(outputs) > 1 else outputs[0]

    def forward(self, xs):
        """Create with inherited class
        """
        raise NotImplementedError()

    def backward(self, gys):
        """Create with inherited class
        """
        raise NotImplementedError()


def as_variable(obj):
    """Convert kdezero.Variable

    Args:
        obj (kdezero.Variable, ndarray):
            Object to convert.
            If obj is kdezero.Variable return it as it is
    
    Returns:
        kdezero.Variable: Variable converted from obj.
    """
    if isinstance(obj, Variable):
        return obj
    return Variable(obj)


def as_array(x, array_module=np):
    """Convert ndarray

    Args:
        x (scalar or ndarray):
            Convert scalar to ndarray, otherwise return as is.
        array_module (np or cp, optional):
            Specify a numpy or cupy module. (default: np)

    Returns:
        ndarray: Array of specified modules
    """
    if np.isscalar(x):
        return array_module.array(x)
    return x



from kdezero import Function
from kdezero import cuda


class Sigmoid(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = xp.tanh(x * 0.5) * 0.5 + 0.5
        return y

    def backward(self, gy):
        y = self.outputs[0]()
        gx = gy * y * (1 - y)
        return gx


def sigmoid(x):
    return Sigmoid()(x)


class ReLU(Function):
    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = xp.maximum(x, 0.0)
        return y

    def backward(self, gy):
        x, = self.inputs
        mask = x.data > 0
        gx = gy * mask
        return gx


def relu(x):
    return ReLU()(x)


class Softmax(Function):
    def __init__(self, axis=1):
        self.axis = axis

    def forward(self, x):
        xp = cuda.get_array_module(x)
        y = x - x.max(axis=self.axis, keepdims=True)
        y = xp.exp(y)
        y /= y.sum(axis=self.axis, keepdims=True)
        return y

    def backward(self, gy):
        y = self.outputs[0]()
        gx = y * gy
        sumdx = gx.sum(axis=self.axis, keepdims=True)
        gx -= y * sumdx
        return gx


def softmax(x, axis=1):
    return Softmax(axis)(x)



from kdezero.datasets import Dataset
from kdezero.utils import get_file


class ImageNet(Dataset):
    """ImageNet labels
    Convert numbers from 0 to 999 to ImageNet labels.
    """
    def __init__(self):
        NotImplemented

    @staticmethod
    def labels():
        """Return labels

        Returns:
            Dict: key is a number from 0 to 999 and value is label.
        """
        url = 'https://gist.githubusercontent.com/yrevar/942d3a0ac09ec9e5eb3a/raw/238f720ff059c1f82f368259d1ca4ffa5dd8f9f5/imagenet1000_clsidx_to_labels.txt'
        path = get_file(url)
        with open(path, 'r') as f:
            labels = eval(f.read())
        return labels



def get_deconv_outsize(size, kernel_size, stride, pad):
    return stride * (size - 1) + kernel_size - 2 + pad


def get_conv_outsize(input_size, kernel_size, stride, pad):
    return (input_size + pad * 2 - kernel_size) // stride + 1


def pair(x):
    if isinstance(x, int):
        return (x, x)
    elif isinstance(x, tuple):
        assert len(x) == 2
        return x
    else:
        raise ValueError



from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setup(
    name="kdezero",
    version='1.7',
    description='This library is an my improved version of "Deep Learning from Basic 3".',
    long_description = long_description,
    long_description_content_type = "text/markdown",
    author='Kota Suzuki',
    author_email='suzuki.kota0331@gmail.com',
    url='https://github.com/kotabrog/K_DeZero.git',
    license='MIT',
    packages=find_packages(),
    install_requires=open('requirements.txt').read().splitlines(),
)



from kdezero.core import Variable
from kdezero.core import Parameter
from kdezero.core import Function
from kdezero.core import using_config
from kdezero.core import no_grad
from kdezero.core import test_mode
from kdezero.core import as_array
from kdezero.core import as_variable
from kdezero.core import setup_variable
from kdezero.core import Config
from kdezero.layers import Layer
from kdezero.models import Model
from kdezero.datasets import Dataset
from kdezero.dataloaders import DataLoader
from kdezero.dataloaders import SeqDataLoader

import kdezero.datasets
import kdezero.dataloaders
import kdezero.optimizers
import kdezero.functions
import kdezero.layers
import kdezero.utils
import kdezero.cuda
import kdezero.transforms

setup_variable()



from kdezero.datasets.core import Dataset
from kdezero.datasets.imagenet import ImageNet
from kdezero.datasets.mnist import MNIST
from kdezero.datasets.sincurve import SinCurve
from kdezero.datasets.spiral import Spiral



import math
from kdezero import cuda

class Optimizer:
    def __init__(self):
        self.target = None
        self.hooks = []

    def setup(self, target):
        self.target = target
        return self

    def update(self):
        params = [p for p in self.target.params() if p.grad is not None]

        for f in self.hooks:
            f(params)

        for param in params:
            self.update_one(param)

    def update_one(self, param):
        raise NotImplementedError()

    def add_hook(self, f):
        self.hooks.append(f)


class SGD(Optimizer):
    def __init__(self, lr=0.01):
        super().__init__()
        self.lr = lr

    def update_one(self, param):
        param.data -= self.lr * param.grad.data


class MomentumSGD(Optimizer):
    def __init__(self, lr=0.01, momentum=0.9):
        super().__init__()
        self.lr = lr
        self.momentum = momentum
        self.vs = {}

    def update_one(self, param):
        v_key = id(param)
        if v_key not in self.vs:
            xp = cuda.get_array_module(param.data)
            self.vs[v_key] = xp.zeros_like(param.data)
        
        v = self.vs[v_key]
        v *= self.momentum
        v -= self.lr * param.grad.data
        param.data += v


class Adam(Optimizer):
    def __init__(self, alpha=0.001, beta1=0.9, beta2=0.999, eps=1e-8):
        super().__init__()
        self.t = 0
        self.alpha = alpha
        self.beta1 = beta1
        self.beta2 = beta2
        self.eps = eps
        self.ms = {}
        self.vs = {}

    def update(self, *args, **kwargs):
        self.t += 1
        super().update(*args, **kwargs)

    @property
    def lr(self):
        fix1 = 1. - math.pow(self.beta1, self.t)
        fix2 = 1. - math.pow(self.beta2, self.t)
        return self.alpha * math.sqrt(fix2) / fix1

    def update_one(self, param):
        xp = cuda.get_array_module(param.data)

        key = id(param)
        if key not in self.ms:
            self.ms[key] = xp.zeros_like(param.data)
            self.vs[key] = xp.zeros_like(param.data)

        m, v = self.ms[key], self.vs[key]
        beta1, beta2, eps = self.beta1, self.beta2, self.eps
        grad = param.grad.data

        m += (1 - beta1) * (grad - m)
        v += (1 - beta2) * (grad * grad - v)
        param.data -= self.lr * m / (xp.sqrt(v) + eps)


class WeightDecay:
    def __init__(self, rate):
        self.rate = rate

    def __call__(self, params):
        for param in params:
            param.grad.data += self.rate * param.data
